![Superalgos video capture (16)](https://user-images.githubusercontent.com/9479367/77251218-76d25980-6c4d-11ea-8e47-be7db2e8abdb.gif)

# What is Superalgos?

Superalgos is a platform to automate crypto-trading. It is implemented as a Nodejs Backend + Web App that runs on your hardware and scales from a single Raspberry Pi to a Trading Farm. Superalgos is **Free** and **Open Source**.

## Superalgos Features

* A Visual Scripting Designer.
* Integrated Charting System.
* A Visual Strategy Debugger.
* Task Management across a Trading Farm.
* Community built strategies to learn and start from.

## Superalgos allows you to

* Visually design your own trading strategies.
* Visually debug your trading strategies.
* Visually design your own indicators.
* Visually design your own plotters (to visualize indicators or data mined).
* Visually design your data-mining operations.
* Download historical market data from crypto exchanges.
* Backtest your strategies against historical data.
* Run live trading sessions.
* Run arbitrary data-mining operations of any size.
* Feed your trading strategies with the data mined.

## Superalgos Development Pipeline

* Ethereum Integration, which will allow you to data mine your own Ethereum node and use the mined data in your strategies.
* Machine Learning, which will allow you to run a Reinforcement Learning Algorithm and feed it with mined data so as to learn how to take decisions relevant for your strategies.

## Superalgos is user-centric

* No ads, anywhere.
* No sign up / logins.
* No user / usage data collection of any kind.
* Runs 100% on uncompiled code anyone can read and audit.

# Superalgos for Developers

* You are free to customize Superalgos for your customers. No royalties, no license fees.
* Everything extra your customers might need you can code it yourself or request it as new features.
* You can use Superalgos as a platform or a component of a larger system.
* No proprietary code / libraries. All open source and free.
* Already available a set of community-contributed plugins (workspaces, strategies, indicators, plotters, tutorials, etc.).

## Superalgos saves you time

* No need to code the download of historical data from crypto exchanges.
* No need to code the streaming of market data from crypto exchanges.
* No need to hardcode strategies, use the visual designer for a more flexible approach.
* No need to debug what went wrong line by line or diving into log files with tons of data. You can see each variable of the state of the Trading Engine at every candle by hovering the mouse over the charts.
* No need to integrate a charting library, Superalgos has its own integrated Charting System.
* No need to manage task data or execution dependencies by yourself. Superalgos allows you to define Tasks and distribute them across a Trading Farm, and it takes care of the data and execution dependencies, so that each task automatically starts when their dependencies are ready.

# Superalgos for Individuals

* Superalgos is easy to install / uninstall.
* Superalgos is easy to run.
* Superalgos is easy to use.
* Superalgos is easy to learn.
* Superalgos is easy to debug.
* Superalgos is well documented.
* You have free online support via Telegram.

## Superalgos saves you money

* There are no paid plans or anything that costs you money.
* There is no locked functionality, you can use the full capacity of the software.
* There is no limit in the number of backtests you can run. 
* There is no limit in the number of live sessions you can run.
* There is no limit in the number of historical data you can download.
* There is no limit in the number of data you can process.
* You can use all the plugins available (indicators / plotters / strategies / etc.)
* You can install Superalgos in as many machines as you want.
* Your installations can be used by as many people as you want.
* You can connect to as many crypto exchanges as you want.

## Superalgos minimizes your risks

* No one can know what strategies you design / run.
* No one can front-run you.
* No one can steal your trading ideas.
* No one knows how much money you have / trade.
* No one can see your exchange keys.

# Superalgos for Companies

* No need to buy expensive software for monitoring crypto markets or trading execution.
* All your employees can use Superalgos for free.
* You can use Superalgos at its full capacity or just the features you are currently interested in.
* Superalgos could be integrated into your existing operation, feeding itself from other systems or feeding other systems as well.

# Getting Started

All procedures are the same for Windows, Linux or Mac OS. Note: only tested with Google Chrome.

## Installation

### Pre-Requisites

1. If you don't have it yet, [download and install Node.js.](https://nodejs.org/en/download/)
2. If you don't have it yet, [download and install Git.](https://git-scm.com/downloads)

### App Setup

1. Open a command prompt or terminal as an administrator and type:
```
git clone https://github.com/Superalgos/Superalgos
```

## Usage

1. Change to the Superalgos directory / folder and type:
```
node run
```
2. Follow the in-app Welcome Tutorial to learn the basics.

## Uninstall

Just delete the folder Superalgos.

# Support

### Via Telegram

Users online support through our [Superalgos Telegram Community](https://t.me/superalgoscommunity).

Developers online support through our [Superalgos Telegram Developer's Group](https://t.me/superalgosdevelop).

Support in Spanish through our [Superalgos en Español Telegram Group](https://t.me/superalgos_es).
 
### Via Documentation Site

Self service support available via the [Superalgos Documentation Site](https://docs.superalgos.org/).

### Via Video Tutorials

Self service support via video tutorials at the [Superalgos YouTube Channel](https://www.youtube.com/channel/UCmYSGbB151xFQPNxj7KfKBg).

# Other Resources

### Web Site

For an overview of what Superalgos can do for you, please check the [Superalgos Website](https://superalgos.org/).

### Blog

You can find everything we have written at the [Superalgos Blog](https://medium.com/superalgos).

### Twitter

To stay up-to-date, follow [Superalgos on Twitter](https://twitter.com/superalgos).

### Facebook

To stay up-to-date, follow [Superalgos on Facebook](https://www.facebook.com/superalgos).

# Contributing

Superalgos is a Community Project built by Contributors. Learn [how to become a Contributor](https://docs.superalgos.org/contributing-to-superalgos.html).

## Contributors 

Luis Fernando Molina, Julian Molina, Andreja Cobeljic, Matías Benitez, Ira Miller, Eduardo Remis, Jeff Braun, rico4dev, 9808us, Barry Low, Nikola Bjelogrlic, Hirajin Koizuko, Francisco J. Santillán, Viktoria B., Guillermo V., Daniel J., Javier A., Gustavo J., Romina GS, Pedro P., Thais M., Andrey M., Loui M., Natalia M., Bashar A., Carlos V., Diego M., Sebastian E., Bogdan P., Marko V., Igor S., Niksa K., Rodrigo M., Nicanor M., Alejandro P., Mateo H., Lan T., Leon A., Uroš R., Filip M., Vladimir J. and Pavle B.

# About this Repository

This repository is the one that stores the output of the build process and it is the one that users actually download to their machines and run. As a system, Superalgos is distributed across different repositories that are combined during the build process to produce the content of this one. A secondary use case for this repo is as issue tracker for the overall project. 

# License

Superalgos is open-source software released under [Apache License 2.0](LICENSE).




